# pyflyby/test_importdb.py

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/



import os
import sys
from   shutil                   import rmtree
from   tempfile                 import NamedTemporaryFile, mkdtemp
from   textwrap                 import dedent

from   pyflyby._importclns      import ImportMap, ImportSet
from   pyflyby._importdb        import ImportDB
from   pyflyby._importstmt      import Import
from   pyflyby._util            import EnvVarCtx

from contextlib import contextmanager


if sys.version_info > (3, 11):
    from contextlib import chdir

else:

    @contextmanager
    def chdir(path):
        old = os.getcwd()
        try:
            os.chdir(path)
            yield
        finally:
            os.chdir(old)


def test_importDB_root():
    """
    See #362
    """
    with chdir("/"):
        ImportDB.get_default(None)

def test_ImportDB_from_code_1():
    db = ImportDB('from aa.bb import cc as dd, ee')
    expected_known = ImportSet(['from aa.bb import cc as dd, ee'])
    assert db.known_imports == expected_known


def test_ImportDB_from_code_complex_1():
    result = ImportDB('''
        import foo, bar as barf
        from xx import yy, yyy, yyyy
        __mandatory_imports__ = ['__future__.division',
                                 'import aa . bb . cc as dd']
        __forget_imports__ = ['xx.yy', 'from xx import zz']
        __canonical_imports__ = {'bad.baad': 'good.goood'}
    ''')
    assert result.known_imports == ImportSet([
        "import foo",
        "import bar as barf",
        "from xx import yyy, yyyy"
        ])
    assert result.mandatory_imports == ImportSet([
        "from __future__ import division",
        "from aa.bb import cc as dd"])
    assert result.forget_imports == ImportSet([
        "from xx import yy",
        "from xx import zz"])
    assert result.canonical_imports == ImportMap({
        "bad.baad": "good.goood"
    })


def test_ImportDB_by_fullname_or_import_as_1():
    db = ImportDB('from aa.bb import cc as dd')
    result = db.by_fullname_or_import_as
    expected = {
        'aa': (Import('import aa'),),
        'aa.bb': (Import('import aa.bb'),),
        'dd': (Import('from aa.bb import cc as dd'),)
    }
    assert result == expected


def test_ImportDB_get_default_1():
    db = ImportDB.get_default('.')
    assert isinstance(db, ImportDB)
    assert ImportDB(db) is db


def import_ImportDB_memoized_1():
    db1 = ImportDB.get_default('.')
    db2 = ImportDB.get_default('.')
    assert db1 is db2


def test_ImportDB_pyflyby_path_filename_1():
    # Check that PYFLYBY_PATH set to a filename works.
    with NamedTemporaryFile(mode='w+') as f:
        f.write("from m4065635 import f78841936, f44111337, f73485346\n")
        f.flush()
        with EnvVarCtx(PYFLYBY_PATH=f.name):
            db = ImportDB.get_default('/bin')
        assert isinstance(db, ImportDB)
        result = db.by_fullname_or_import_as["f44111337"]
        expected = (Import('from m4065635 import f44111337'),)
        assert result == expected


def test_ImportDB_pyflyby_path_no_default_1():
    # Check that defaults can be turned off from PYFLYBY_PATH.
    with NamedTemporaryFile(mode='w+') as f:
        f.write("from m27056973 import f8855924\n")
        f.flush()
        with EnvVarCtx(PYFLYBY_PATH=f.name):
            db = ImportDB.get_default('/bin')
        assert isinstance(db, ImportDB)
        result = db.by_fullname_or_import_as["f8855924"]
        expected = (Import('from m27056973 import f8855924'),)
        assert result == expected
        assert "defaultdict" not in db.by_fullname_or_import_as
        expected_bfoia = {
            "f8855924": (Import("from m27056973 import f8855924"),),
            "m27056973": (Import("import m27056973"),),
        }
        assert db.by_fullname_or_import_as == expected_bfoia
    # For the default PYFLYBY_PATH (configured in conftest.py), we should have
    # defaultdict.
    db2 = ImportDB.get_default('/bin')
    result = db2.by_fullname_or_import_as["defaultdict"]
    expected = (Import('from collections import defaultdict'),)
    assert result == expected


def test_ImportDB_pyflyby_path_change_1():
    # Check that memoization takes into account changes in
    # os.environ["PYFLYBY_PATH"].
    with NamedTemporaryFile(mode='w+') as f:
        f.write("from m60309242 import f5781152\n")
        f.flush()
        with EnvVarCtx(PYFLYBY_PATH=f.name):
            db = ImportDB.get_default('/bin')
        result = db.by_fullname_or_import_as["f5781152"]
        expected = (Import('from m60309242 import f5781152'),)
        assert result == expected
        db2 = ImportDB.get_default('/bin')
        assert db2 is not db
        assert "f5781152" not in db2.by_fullname_or_import_as


def test_ImportDB_pyflyby_recurse_dir_1():
    d = mkdtemp(prefix=".", suffix="_pyflyby")
    with EnvVarCtx(PYFLYBY_PATH=d):
        os.mkdir("%s/d1"%d)
        os.mkdir("%s/d1/d2"%d)
        os.mkdir("%s/d1/d2/d3"%d)
        os.mkdir("%s/.d4"%d)
        with open("%s/d1/d2/d3/f6446612.py"%d, 'w') as f:
            f.write("from m7540535 import f17684046, f7241844")
        with open("%s/d1/d2/d3/f91456848"%d, 'w') as f: # missing ".py"
            f.write("from m5733351 import f17684046, f7241844")
        with open("%s/.d4/f52247912.py"%d, 'w') as f: # under a dot dir
            f.write("from m50938634 import f17684046, f7241844")
        db = ImportDB.get_default("/bin")
        result = db.by_fullname_or_import_as["f7241844"]
        expected = (Import("from m7540535 import f7241844"),)
        assert result == expected
        rmtree(d)


def test_ImportDB_pyflyby_dotdotdot_1():
    with EnvVarCtx(PYFLYBY_PATH=".../f1198375"):
        d = mkdtemp("_pyflyby")
        os.mkdir("%s/d1"%d)
        os.mkdir("%s/d1/d2"%d)
        os.mkdir("%s/d1/d2/d3"%d)
        with open("%s/f1198375"%d, 'w') as f:
            f.write("from m97722423 import f49463937, f3532073\n")
        with open("%s/d1/d2/f1198375"%d, 'w') as f:
            f.write("from m90927291 import f6273971, f49463937\n")
        db = ImportDB.get_default("%s/d1/d2/d3/f"%d)
        result = db.by_fullname_or_import_as["f49463937"]
        expected = (
            Import("from m90927291 import f49463937"),
            Import("from m97722423 import f49463937"),
        )
        assert result == expected
        rmtree(d)


def test_ImportDB_pyflyby_forget_1():
    with EnvVarCtx(PYFLYBY_PATH=".../f70301376"):
        d = mkdtemp("_pyflyby")
        os.mkdir("%s/d1"%d)
        os.mkdir("%s/d1/d2"%d)
        os.mkdir("%s/d1/d2/d3"%d)
        with open("%s/f70301376"%d, 'w') as f:
            f.write(dedent("""
                from m49790901       import f27626336, f96186952
                from m78687343       import f35156295, f43613649
                from m78687343.a.b.c import f54583581
                from m49790901.a     import f27626336, f96186952
            """))
        with open("%s/d1/d2/f70301376"%d, 'w') as f:
            f.write(dedent("""
                __forget_imports__ = [
                   'from m49790901 import f27626336',
                   'from m78687343 import *',
                ]
            """))
        db = ImportDB.get_default("%s/d1/d2/d3/f"%d)
        result = db.known_imports
        expected = ImportSet("""
                from m49790901       import f96186952
                from m49790901.a     import f27626336, f96186952
        """)
        assert result == expected
        rmtree(d)



import matplotlib
from   matplotlib               import pyplot
import matplotlib.colors
from   matplotlib.colors        import ColorConverter
from   matplotlib.font_manager  import FontProperties
from   matplotlib.patches       import Rectangle
from   matplotlib.pyplot        import (clf, draw, figure, gca, gcf, ioff,
                                        legend, plot, savefig, scatter, show,
                                        subplot, title, xlabel, ylabel, ylim)
from   matplotlib.ticker        import (Formatter, Locator, NullFormatter,
                                        NullLocator)
import numexpr
import numpy as np, numpy as npy, numpy
from   numpy                    import (Inf, NAN, NaN, abs as aabs, absolute,
                                        add, all as aall, allclose, alltrue,
                                        amax, amin, angle, any as aany,
                                        append as aappend, apply_along_axis,
                                        apply_over_axes, arange, arccos,
                                        arccosh, arcsin, arcsinh, arctan,
                                        arctan2, arctanh, argmax, argmin,
                                        argsort, argwhere, around, array,
                                        array2string, array_equal, array_equiv,
                                        array_repr, array_split, array_str,
                                        asanyarray, asarray, asarray_chkfinite,
                                        ascontiguousarray, asfarray,
                                        asfortranarray, asmatrix, asscalar,
                                        atleast_1d, atleast_2d, atleast_3d,
                                        average, bartlett, base_repr,
                                        binary_repr, bincount, bitwise_and,
                                        bitwise_not, bitwise_or, bitwise_xor,
                                        blackman, bmat, bool8, bool_,
                                        broadcast, broadcast_arrays, byte,
                                        byte_bounds, c_, can_cast, cdouble,
                                        ceil, cfloat, character, chararray,
                                        choose, clip, clongdouble, clongfloat,
                                        column_stack, common_type,
                                        compare_chararrays, compat, complex128,
                                        complex64, complex_, complexfloating,
                                        concatenate, conj, conjugate, convolve,
                                        copy, copysign, corrcoef, correlate,
                                        cos, cosh, cov, cross, csingle,
                                        ctypeslib, cumprod, cumproduct, cumsum,
                                        deg2rad, degrees, diag, diag_indices,
                                        diag_indices_from, diagflat, diagonal,
                                        diff, digitize, disp, divide, dot,
                                        double, dsplit, dstack, dtype, ediff1d,
                                        einsum, emath, empty, empty_like,
                                        equal, exp, exp2, expand_dims, expm1,
                                        extract, eye, fabs,
                                        fastCopyAndTranspose, fill_diagonal,
                                        find_common_type, fix, flatiter,
                                        flatnonzero, fliplr, flipud, float32,
                                        float64, float_, floating, floor,
                                        floor_divide, fmax, fmin, fmod, frexp,
                                        frombuffer, fromfile, fromfunction,
                                        fromiter, frompyfunc, fromregex,
                                        fromstring, gradient, greater,
                                        greater_equal, hamming, hanning,
                                        heaviside, histogram, histogram2d,
                                        histogramdd, hsplit, hstack, hypot,
                                        i0, identity, iinfo, imag, in1d,
                                        index_exp, indices, inexact, inf,
                                        inner, int0, int16, int32, int64,
                                        int8, int_, intc, integer, interp,
                                        intersect1d, intp, invert, ipmt, irr,
                                        iscomplex, iscomplexobj, isfinite,
                                        isfortran, isinf, isnan, isneginf,
                                        isposinf, isreal, isrealobj, isscalar,
                                        issctype, issubclass_, issubdtype,
                                        issubsctype, iterable, ix_, kaiser,
                                        kron, ldexp, left_shift, less,
                                        less_equal, lexsort, linalg, linspace,
                                        little_endian, loadtxt, log,
                                        log as logarithm, log10, log1p, log2,
                                        logaddexp, logaddexp2, logical_and,
                                        logical_not, logical_or, logical_xor,
                                        logspace, longcomplex, longdouble,
                                        longfloat, longlong, mafromtxt,
                                        mask_indices, mat, matrix, maximum,
                                        mean, median, memmap, meshgrid, mgrid,
                                        minimum, mintypecode, mirr, mod, modf,
                                        msort, multiply, nan, nan_to_num,
                                        nanargmax, nanargmin, nanmax, nanmin,
                                        nansum, nbytes, ndarray, ndenumerate,
                                        ndim, ndindex, negative, newaxis,
                                        newbuffer, nextafter, nonzero,
                                        not_equal, nper, npv, number, object0,
                                        object_, ogrid, ones, ones_like,
                                        outer, packbits, pi, piecewise,
                                        pkgload, place, pmt, poly, poly1d,
                                        polyadd, polyder, polydiv, polyfit,
                                        polyint, polymul, polynomial, polysub,
                                        polyval, power, ppmt, prod, product,
                                        ptp, putmask, pv, r_, rad2deg,
                                        radians, rank, rate, ravel, real,
                                        real_if_close, recarray, recfromcsv,
                                        recfromtxt, reciprocal, record,
                                        remainder, repeat, reshape, resize,
                                        restoredot, right_shift, rint, roll,
                                        rollaxis, roots, rot90, round, round_,
                                        row_stack, s_, searchsorted,
                                        select as aselect, setbufsize,
                                        setdiff1d, setxor1d, shape, short,
                                        show_config, sign, signbit,
                                        signedinteger, sin, sinc, single,
                                        singlecomplex, sinh, size as asize,
                                        sometrue, sort as asort, sort,
                                        sort_complex, spacing, split, sqrt,
                                        square, squeeze, std, str_, string0,
                                        string_, subtract, sum, swapaxes, take,
                                        tan, tanh, tensordot, testing, tile,
                                        trace, transpose, trapz, tri, tril,
                                        tril_indices, tril_indices_from,
                                        trim_zeros, triu, triu_indices,
                                        triu_indices_from, true_divide, trunc,
                                        ubyte, ufunc, uint, uint0, uint16,
                                        uint32, uint64, uint8, uintc, uintp,
                                        ulonglong, unicode0, unicode_, union1d,
                                        unique, unpackbits, unravel_index,
                                        unsignedinteger, unwrap, ushort,
                                        vander, var, vdot, vectorize, void,
                                        void0, vsplit, vstack, where, zeros,
                                        zeros_like)
from   numpy.core.umath_tests   import inner1d
from   numpy.fft                import (fft, fft2, fftn, ifft, ifft2, ifftn,
                                        irfft, irfft2, irfftn, rfft, rfft2,
                                        rfftn)
from   numpy.lib                import recfunctions as recf, recfunctions
from   numpy.lib.stride_tricks  import as_strided
import numpy.linalg
from   numpy.linalg             import cholesky, det, eigh, inv, pinv, svd
from   numpy.random             import (normal, rand, randint, randn,
                                        random as arandom, shuffle)
import numpy.testing
import numpy.version
import pandas, pandas as pd
from   pandas                   import DataFrame, Series, TimeSeries
import pylab as pl, pylab
import scipy
from   scipy                    import integrate, optimize, special, stats
import scipy.cluster.hierarchy
import scipy.integrate
import scipy.interpolate
from   scipy.interpolate        import InterpolatedUnivariateSpline, interp1d
import scipy.linalg
import scipy.optimize
from   scipy.optimize           import (curve_fit, fmin_l_bfgs_b, fsolve,
                                        leastsq)
from   scipy.optimize.zeros     import bisect
import scipy.special
from   scipy.special            import gamma, gammainc, gammaincinv, ndtri
import scipy.stats
from   scipy.stats              import (chisqprob, distributions,
                                        scoreatpercentile, uniform)
from   scipy.stats.distributions \
                                import norm



# pyflyby/_cmdline.py.
# Copyright (C) 2011, 2012, 2013, 2014, 2015, 2018 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT



from   builtins                 import input
import optparse
import os
import signal
import sys
from   textwrap                 import dedent
import traceback
from   typing                   import List


from   pyflyby._file            import (FileText, Filename, atomic_write_file,
                                        expand_py_files_from_args, read_file)
from   pyflyby._importstmt      import ImportFormatParams
from   pyflyby._log             import logger
from   pyflyby._util            import cached_attribute, indent


def hfmt(s):
    return dedent(s).strip()

def maindoc():
    import __main__
    return (__main__.__doc__ or '').strip()


def _sigpipe_handler(*args):
    # The parent process piped our stdout and closed the pipe before we
    # finished writing, e.g. "tidy-imports ... | head" or "tidy-imports ... |
    # less".  Exit quietly - squelch the "close failed in file object
    # destructor" message would otherwise be raised.
    raise SystemExit(1)


def parse_args(
    addopts=None, import_format_params=False, modify_action_params=False, defaults=None
):
    """
    Do setup for a top-level script and parse arguments.
    """

    if defaults is None:
        defaults = {}

    ### Setup.
    # Register a SIGPIPE handler.
    signal.signal(signal.SIGPIPE, _sigpipe_handler)
    ### Parse args.
    parser = optparse.OptionParser(usage='\n'+maindoc())

    def log_level_callbacker(level):
        def callback(option, opt_str, value, parser):
            logger.set_level(level)
        return callback

    def debug_callback(option, opt_str, value, parser):
        logger.set_level("DEBUG")

    parser.add_option("--debug", action="callback",
                      callback=debug_callback,
                      help="Debug mode (noisy and fail fast).")

    parser.add_option("--verbose", action="callback",
                      callback=log_level_callbacker("DEBUG"),
                      help="Be noisy.")

    parser.add_option("--quiet", action="callback",
                      callback=log_level_callbacker("ERROR"),
                      help="Be quiet.")

    parser.add_option("--version", action="callback",
                      callback=lambda *args: print_version_and_exit(),
                      help="Print pyflyby version and exit.")

    if modify_action_params:
        group = optparse.OptionGroup(parser, "Action options")
        action_diff = action_external_command('pyflyby-diff')
        def parse_action(v):
            V = v.strip().upper()
            if V == 'PRINT':
                return action_print
            elif V == 'REPLACE':
                return action_replace
            elif V == 'QUERY':
                return action_query()
            elif V == "DIFF":
                return action_diff
            elif V.startswith("QUERY:"):
                return action_query(v[6:])
            elif V.startswith("EXECUTE:"):
                return action_external_command(v[8:])
            elif V == "IFCHANGED":
                return action_ifchanged
            elif V == "EXIT1":
                return action_exit1
            else:
                raise Exception(
                    "Bad argument %r to --action; "
                    "expected PRINT or REPLACE or QUERY or IFCHANGED or EXIT1 "
                    "or EXECUTE:..." % (v,))

        def set_actions(actions):
            actions = tuple(actions)
            parser.values.actions = actions

        def action_callback(option, opt_str, value, parser):
            action_args = value.split(',')
            set_actions([parse_action(v) for v in action_args])

        def action_callbacker(actions):
            def callback(option, opt_str, value, parser):
                set_actions(actions)
            return callback

        group.add_option(
            "--actions", type='string', action='callback',
            callback=action_callback,
            metavar='PRINT|REPLACE|IFCHANGED|QUERY|DIFF|EXIT1:EXECUTE:mycommand',
            help=hfmt('''
                   Comma-separated list of action(s) to take.  If PRINT, print
                   the changed file to stdout.  If REPLACE, then modify the
                   file in-place.  If EXECUTE:mycommand, then execute
                   'mycommand oldfile tmpfile'.  If DIFF, then execute
                   'pyflyby-diff'.  If QUERY, then query user to continue.
                   If IFCHANGED, then continue actions only if file was
                   changed.  If EXIT1, then exit with exit code 1 after all
                   files/actions are processed.'''))
        group.add_option(
            "--print", "-p", action='callback',
            callback=action_callbacker([action_print]),
            help=hfmt('''
                Equivalent to --action=PRINT (default when stdin or stdout is
                not a tty) '''))
        group.add_option(
            "--diff", "-d", action='callback',
            callback=action_callbacker([action_diff]),
            help=hfmt('''Equivalent to --action=DIFF'''))
        group.add_option(
            "--replace", "-r", action='callback',
            callback=action_callbacker([action_ifchanged, action_replace]),
            help=hfmt('''Equivalent to --action=IFCHANGED,REPLACE'''))
        group.add_option(
            "--diff-replace", "-R", action='callback',
            callback=action_callbacker([action_ifchanged, action_diff, action_replace]),
            help=hfmt('''Equivalent to --action=IFCHANGED,DIFF,REPLACE'''))
        actions_interactive = [
            action_ifchanged, action_diff,
            action_query("Replace {filename}?"), action_replace]
        group.add_option(
            "--interactive", "-i", action='callback',
            callback=action_callbacker(actions_interactive),
            help=hfmt('''
               Equivalent to --action=IFCHANGED,DIFF,QUERY,REPLACE (default
               when stdin & stdout are ttys) '''))
        if os.isatty(0) and os.isatty(1):
            default_actions = actions_interactive
        else:
            default_actions = [action_print]
        parser.set_default('actions', tuple(default_actions))
        parser.add_option_group(group)

        parser.add_option(
            '--symlinks', action='callback', nargs=1, type=str,
            dest='symlinks', callback=symlink_callback, help="--symlinks should be one of: " + symlinks_help,
        )
        parser.set_defaults(symlinks='error')

    if import_format_params:
        group = optparse.OptionGroup(parser, "Pretty-printing options")
        group.add_option('--align-imports', '--align', type='str', default="32",
                         metavar='N',
                         help=hfmt('''
                             Whether and how to align the 'import' keyword in
                             'from modulename import aliases...'.  If 0, then
                             don't align.  If 1, then align within each block
                             of imports.  If an integer > 1, then align at
                             that column, wrapping with a backslash if
                             necessary.  If a comma-separated list of integers
                             (tab stops), then pick the column that results in
                             the fewest number of lines total per block.'''))
        group.add_option('--from-spaces', type='int', default=3, metavar='N',
                         help=hfmt('''
                             The number of spaces after the 'from' keyword.
                             (Must be at least 1; default is 3.)'''))
        group.add_option('--separate-from-imports', action='store_true',
                         default=False,
                         help=hfmt('''
                             Separate 'from ... import ...'
                             statements from 'import ...' statements.'''))
        group.add_option('--no-separate-from-imports', action='store_false',
                         dest='separate_from_imports',
                         help=hfmt('''
                            (Default) Don't separate 'from ... import ...'
                            statements from 'import ...' statements.'''))
        group.add_option('--align-future', action='store_true',
                         default=False,
                         help=hfmt('''
                             Align the 'from __future__ import ...' statement
                             like others.'''))
        group.add_option('--no-align-future', action='store_false',
                         dest='align_future',
                         help=hfmt('''
                             (Default) Don't align the 'from __future__ import
                             ...' statement.'''))
        group.add_option('--width', type='int', default=None, metavar='N',
                         help=hfmt('''
                             Maximum line length (default: 79).'''))
        group.add_option('--black', action='store_true', default=False,
                         help=hfmt('''
                             Use black to format imports. If this option is
                             used, all other formatting options are ignored,
                             except width'''))
        group.add_option('--hanging-indent', type='choice', default='never',
                         choices=['never','auto','always'],
                         metavar='never|auto|always',
                         dest='hanging_indent',
                         help=hfmt('''
                             How to wrap import statements that don't fit on
                             one line.
                             If --hanging-indent=always, then always indent
                             imported tokens at column 4 on the next line.
                             If --hanging-indent=never (default), then align
                             import tokens after "import (" (by default column
                             40); do so even if some symbols are so long that
                             this would exceed the width (by default 79)).
                             If --hanging-indent=auto, then use hanging indent
                             only if it is necessary to prevent exceeding the
                             width (by default 79).
                         '''))
        def uniform_callback(option, opt_str, value, parser):
            parser.values.separate_from_imports = False
            parser.values.from_spaces           = 3
            parser.values.align_imports         = '32'
        group.add_option('--uniform', '-u', action="callback",
                         callback=uniform_callback,
                         help=hfmt('''
                             (Default) Shortcut for --no-separate-from-imports
                             --from-spaces=3 --align-imports=32.'''))
        def unaligned_callback(option, opt_str, value, parser):
            parser.values.separate_from_imports = True
            parser.values.from_spaces           = 1
            parser.values.align_imports         = '0'
        group.add_option('--unaligned', '-n', action="callback",
                         callback=unaligned_callback,
                         help=hfmt('''
                             Shortcut for --separate-from-imports
                             --from-spaces=1 --align-imports=0.'''))

        parser.add_option_group(group)
    if addopts is not None:
        addopts(parser)
    # This is the only way to provide a default value for an option with a
    # callback.
    if modify_action_params:
        args = ["--symlinks=error"] + sys.argv[1:]
    else:
        args = None
    options, args = parser.parse_args(args=args)
    if import_format_params:
        align_imports_args = [int(x.strip())
                              for x in options.align_imports.split(",")]
        if len(align_imports_args) == 1 and align_imports_args[0] == 1:
            align_imports = True
        elif len(align_imports_args) == 1 and align_imports_args[0] == 0:
            align_imports = False
        else:
            align_imports = tuple(sorted(set(align_imports_args)))
        options.params = ImportFormatParams(
            align_imports         =align_imports,
            from_spaces           =options.from_spaces,
            separate_from_imports =options.separate_from_imports,
            max_line_length       =options.width,
            use_black             =options.black,
            align_future          =options.align_future,
            hanging_indent        =options.hanging_indent,
            )
    return options, args


def _default_on_error(filename):
    raise SystemExit("bad filename %s" % (filename,))


def filename_args(args: List[str], on_error=_default_on_error):
    """
    Return list of filenames given command-line arguments.

    :rtype:
      ``list`` of `Filename`
    """
    if args:
        for a in args:
            assert isinstance(a, str)
        return expand_py_files_from_args([Filename(f) for f in args], on_error)
    elif not os.isatty(0):
        return [Filename.STDIN]
    else:
        syntax()


def print_version_and_exit(extra=None):
    from pyflyby._version import __version__
    msg = "pyflyby %s" % (__version__,)
    progname = os.path.realpath(sys.argv[0])
    if os.path.exists(progname):
        msg += " (%s)" % (os.path.basename(progname),)
    print(msg)
    if extra:
        print(extra)
    raise SystemExit(0)


def syntax(message=None, usage=None):
    if message:
        logger.error(message)
    outmsg = ((usage or maindoc()) +
              '\n\nFor usage, see: %s --help' % (sys.argv[0],))
    print(outmsg, file=sys.stderr)
    raise SystemExit(1)


class AbortActions(Exception):
    pass


class Exit1(Exception):
    pass


class Modifier(object):
    def __init__(self, modifier, filename):
        self.modifier = modifier
        self.filename = filename
        self._tmpfiles = []

    @cached_attribute
    def input_content(self):
        return read_file(self.filename)

    # TODO: refactor to avoid having these heavy-weight things inside a
    # cached_attribute, which causes annoyance while debugging.
    @cached_attribute
    def output_content(self):
        return FileText(self.modifier(self.input_content), filename=self.filename)

    def _tempfile(self):
        from tempfile import NamedTemporaryFile
        f = NamedTemporaryFile()
        self._tmpfiles.append(f)
        return f, Filename(f.name)


    @cached_attribute
    def output_content_filename(self):
        f, fname = self._tempfile()
        f.write(bytes(self.output_content.joined, "utf-8"))
        f.flush()
        return fname

    @cached_attribute
    def input_content_filename(self):
        if isinstance(self.filename, Filename):
            return self.filename
        # If the input was stdin, and the user wants a diff, then we need to
        # write it to a temp file.
        f, fname = self._tempfile()
        f.write(bytes(self.input_content, "utf-8"))
        f.flush()
        return fname


    def __del__(self):
        for f in self._tmpfiles:
            f.close()


def process_actions(filenames:List[str], actions, modify_function,
                    reraise_exceptions=()):
    errors = []
    def on_error_filename_arg(arg):
        print("%s: bad filename %s" % (sys.argv[0], arg), file=sys.stderr)
        errors.append("%s: bad filename" % (arg,))
    filenames = filename_args(filenames, on_error=on_error_filename_arg)
    exit_code = 0
    for filename in filenames:
        try:
            m = Modifier(modify_function, filename)
            for action in actions:
                action(m)
        except AbortActions:
            continue
        except reraise_exceptions:
            raise
        except Exit1:
            exit_code = 1
        except Exception as e:
            errors.append("%s: %s: %s" % (filename, type(e).__name__, e))
            type_e = type(e)
            try:
                tb = sys.exc_info()[2]
                if str(filename) not in str(e):
                    try:
                        e = type_e("While processing %s: %s" % (filename, e))
                        pass
                    except TypeError:
                        # Exception takes more than one argument
                        pass
                if logger.debug_enabled:
                    raise
                traceback.print_exception(type(e), e, tb)
            finally:
                tb = None # avoid refcycles involving tb
            continue
    if errors:
        msg = "\n%s: encountered the following problems:\n" % (sys.argv[0],)
        for er in errors:
            lines = er.splitlines()
            msg += "    " + lines[0] + '\n'.join(
                ("            %s"%line for line in lines[1:]))
        raise SystemExit(msg)
    else:
        raise SystemExit(exit_code)


def action_print(m):
    output_content = m.output_content
    sys.stdout.write(output_content.joined)


def action_ifchanged(m):
    if m.output_content.joined == m.input_content.joined:
        logger.debug("unmodified: %s", m.filename)
        raise AbortActions


def action_replace(m):
    if m.filename == Filename.STDIN:
        raise Exception("Can't replace stdio in-place")
    logger.info("%s: *** modified ***", m.filename)
    atomic_write_file(m.filename, m.output_content)


def action_exit1(m):
    logger.debug("action_exit1")
    raise Exit1


def action_external_command(command):
    import subprocess
    def action(m):
        bindir = os.path.dirname(os.path.realpath(sys.argv[0]))
        env = os.environ
        env['PATH'] = env['PATH'] + ":" + bindir
        fullcmd = "%s %s %s" % (
            command, m.input_content_filename, m.output_content_filename)
        logger.debug("Executing external command: %s", fullcmd)
        ret = subprocess.call(fullcmd, shell=True, env=env)
        logger.debug("External command returned %d", ret)
    return action


def action_query(prompt="Proceed?"):
    def action(m):
        p = prompt.format(filename=m.filename)
        print()
        print("%s [y/N] " % (p), end="")
        try:
            if input().strip().lower().startswith('y'):
                return True
        except KeyboardInterrupt:
            print("KeyboardInterrupt", file=sys.stderr)
            raise SystemExit(1)
        print("Aborted")
        raise AbortActions
    return action

def symlink_callback(option, opt_str, value, parser):
    parser.values.actions = tuple(i for i in parser.values.actions if i not in
    symlink_callbacks.values())
    if value in symlink_callbacks:
        parser.values.actions = (symlink_callbacks[value],) + parser.values.actions
    else:
        raise optparse.OptionValueError("--symlinks must be one of 'error', 'follow', 'skip', or 'replace'. Got %r" % value)

symlinks_help = """\
--symlinks=error (default; gives an error on symlinks),
--symlinks=follow (follows symlinks),
--symlinks=skip (skips symlinks),
--symlinks=replace (replaces symlinks with the target file\
"""

# Warning, the symlink actions will only work if they are run first.
# Otherwise, output_content may already be cached
def symlink_error(m):
    if m.filename == Filename.STDIN:
        return symlink_follow(m)
    if m.filename.islink:
        raise SystemExit("""\
Error: %s appears to be a symlink. Use one of the following options to allow symlinks:
%s
""" % (m.filename, indent(symlinks_help, '    ')))

def symlink_follow(m):
    if m.filename == Filename.STDIN:
        return
    if m.filename.islink:
        logger.info("Following symlink %s" % m.filename)
        m.filename = m.filename.realpath

def symlink_skip(m):
    if m.filename == Filename.STDIN:
        return symlink_follow(m)
    if m.filename.islink:
        logger.info("Skipping symlink %s" % m.filename)
        raise AbortActions

def symlink_replace(m):
    if m.filename == Filename.STDIN:
        return symlink_follow(m)
    if m.filename.islink:
        logger.info("Replacing symlink %s" % m.filename)
        # The current behavior automatically replaces symlinks, so do nothing

symlink_callbacks = {
    'error': symlink_error,
    'follow': symlink_follow,
    'skip': symlink_skip,
    'replace': symlink_replace,
}



# pyflyby/_flags.py.
# Copyright (C) 2011, 2012, 2013, 2014 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT



import __future__
import ast
import operator
from   functools                import reduce
import warnings

from   pyflyby._util            import cached_attribute
from typing import Tuple

# Initialize mappings from compiler_flag to feature name and vice versa.
_FLAG2NAME = {}
_NAME2FLAG = {}
for name in __future__.all_feature_names:
    flag = getattr(__future__, name).compiler_flag
    _FLAG2NAME[flag] = name
    _NAME2FLAG[name] = flag
for name in dir(ast):
    if name.startswith('PyCF'):
        flag_name = name[len('PyCF_'):].lower()
        flag = getattr(ast, name)
        _FLAG2NAME[flag] = flag_name
        _NAME2FLAG[flag_name] = flag
_FLAGNAME_ITEMS = sorted(_FLAG2NAME.items())
_ALL_FLAGS = reduce(operator.or_, _FLAG2NAME.keys())



class CompilerFlags(int):
    """
    Representation of Python "compiler flags", i.e. features from __future__.

      >>> print(CompilerFlags(0x18000).__interactive_display__()) # doctest: +SKIP
      CompilerFlags(0x18000) # from __future__ import with_statement, print_function

      >>> print(CompilerFlags(0x10000, 0x8000).__interactive_display__()) # doctest: +SKIP
      CompilerFlags(0x18000) # from __future__ import with_statement, print_function

      >>> print(CompilerFlags('with_statement', 'print_function').__interactive_display__()) # doctest: +SKIP
      CompilerFlags(0x18000) # from __future__ import with_statement, print_function

      >>> compile("print('x', file=None)", "?", "exec", flags=CompilerFlags("print_function"), dont_inherit=1) #doctest:+ELLIPSIS
      <code object ...>

    """

    # technically both those are compiler flags, but we can't use Self. May need typing_extensions ?
    _ZERO:int
    _UNKNOWN: int

    def __new__(cls, *args):
        """
        Construct a new ``CompilerFlags`` instance.

        :param args:
          Any number (zero or more) ``CompilerFlags`` s, ``int`` s, or ``str`` s,
          which are bitwise-ORed together.
        :rtype:
          `CompilerFlags`
        """
        if len(args) == 0:
            return cls._ZERO
        elif len(args) == 1:
            arg, = args
            if isinstance(arg, cls):
                return arg
            elif arg is None:
                return cls._ZERO
            elif isinstance(arg, int):
                warnings.warn('creating CompilerFlags from integers is deprecated, '
                ' flags values change between Python versions. If you are sure use .from_int',
                DeprecationWarning, stacklevel=2)
                return cls.from_int(arg)
            elif isinstance(arg, str):
                return cls.from_str(arg)
            elif isinstance(arg, ast.AST):
                return cls.from_ast(arg)
            elif isinstance(arg, (tuple, list)):
                return cls(*arg)
            else:
                raise TypeError("CompilerFlags: unknown type %s"
                                % (type(arg).__name__,))
        else:
            flags = []
            for x in args:
                if isinstance(x, cls):
                    flags.append(int(x))
                elif isinstance(x, int):
                    warnings.warn(
                        "creating CompilerFlags from integers is deprecated, "
                        " flags values change between Python versions. If you are sure use .from_int",
                        DeprecationWarning,
                        stacklevel=2,
                    )
                    flags.append(x)
                elif isinstance(x, str):
                    flags.append(int(cls(x)))
                else:
                    raise ValueError

            #assert flags == [0x10000, 0x8000], flags

            return cls.from_int(reduce(operator.or_, flags))

    @classmethod
    def from_int(cls, arg):
        if arg == -1:
            return cls._UNKNOWN  # Instance optimization
        if arg == 0:
            return cls._ZERO # Instance optimization
        self = int.__new__(cls, arg)
        bad_flags = int(self) & ~_ALL_FLAGS
        if bad_flags:
            raise ValueError(
                "CompilerFlags: unknown flag value(s) %s %s" % (bin(bad_flags), hex(bad_flags)))
        return self

    @classmethod
    def from_str(cls, arg:str):
        try:
            flag = _NAME2FLAG[arg]
        except KeyError:
            raise ValueError(
                "CompilerFlags: unknown flag %r" % (arg,))
        return cls.from_int(flag)

    @classmethod
    def from_ast(cls, nodes):
        """
        Parse the compiler flags from AST node(s).

        :type nodes:
          ``ast.AST`` or sequence thereof
        :rtype:
          ``CompilerFlags``
        """
        if isinstance(nodes, ast.Module):
            nodes = nodes.body
        elif isinstance(nodes, ast.AST):
            nodes = [nodes]
        flags = []
        for node in nodes:
            if not isinstance(node, ast.ImportFrom):
                # Got a non-import; stop looking further.
                break
            if not node.module == "__future__":
                # Got a non-__future__-import; stop looking further.
                break
            # Get the feature names.
            names = [n.name for n in node.names]
            flags.extend(names)
        return cls(flags)

    @cached_attribute
    def names(self) -> Tuple[str, ...]:
        return tuple(
            n
            for f, n in _FLAGNAME_ITEMS
            if f & self)

    def __or__(self, o):
        if o == 0:
            return self
        if not isinstance(o, CompilerFlags):
            o = CompilerFlags(o)
        if self == 0:
            return o
        return CompilerFlags.from_int(int(self) | int(o))

    def __ror__(self, o):
        return self | o

    def __and__(self, o):
        if not isinstance(o, int):
            o = CompilerFlags(o)
        return CompilerFlags.from_int(int(self) & int(o))

    def __rand__(self, o):
        return self & o

    def __xor__(self, o):
        if not isinstance(o, CompilerFlags):
            o = CompilerFlags.from_int(o)
        return CompilerFlags.from_int(int(self) ^ int(o))

    def __rxor__(self, o):
        return self ^ o

    def __repr__(self):
        return "CompilerFlags(%s)" % (hex(self),)

    def __str__(self):
        return hex(self)

    def __interactive_display__(self):
        s = repr(self)
        if self != 0:
            s += " # from __future__ import " + ", ".join(self.names)
        return s


CompilerFlags._ZERO = int.__new__(CompilerFlags, 0)
CompilerFlags._UNKNOWN = int.__new__(CompilerFlags, -1)

# flags that _may_ exists on future versions.
_future_flags = {
    "nested_scopes",
    "generators",
    "division",
    "absolute_import",
    "with_statement",
    "print_function",
    "unicode_literals",
    "barry_as_FLUFL",
    "generator_stop",
    "annotations",
    "allow_top_level_await",
    "only_ast",
    "type_comments",
}
for k in _future_flags:
    setattr(CompilerFlags, k, CompilerFlags._UNKNOWN)

for k, v in _NAME2FLAG.items():
    setattr(CompilerFlags, k, CompilerFlags.from_int(v))



# pyflyby/test_flags.py

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/



import ast


import pytest
import warnings

from   pyflyby._flags           import CompilerFlags


def skip_if_new_compiler_flags_values(func):

    return pytest.mark.skipif(
        int(CompilerFlags.print_function) != 0x10000,
        reason="Python flags values changes around Python == 3.8.2",
    )(func)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert int(CompilerFlags(0x18000)) == 0x18000


def test_CompilerFlags_zero_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert CompilerFlags() == CompilerFlags(None) == CompilerFlags(0)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_eqne_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert      CompilerFlags(0x18000) == CompilerFlags(0x18000)
        assert not (CompilerFlags(0x18000) != CompilerFlags(0x18000))
        assert CompilerFlags(0x18000) != CompilerFlags.type_comments
        assert not (CompilerFlags(0x18000) == CompilerFlags(0x10000))


@skip_if_new_compiler_flags_values
def test_CompilerFlags_eqne_other_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert CompilerFlags(0) != object()


@skip_if_new_compiler_flags_values
def test_CompilerFlags_names_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert CompilerFlags(0x18000).names == ("with_statement", "print_function")


@skip_if_new_compiler_flags_values
def test_CompilerFlags_from_int_multi_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert CompilerFlags(0x10000, 0x8000) == CompilerFlags(0x18000)
        assert CompilerFlags(0x10000, 0x8000, 0) == CompilerFlags(0x18000)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_from_CompilerFlags_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        result = CompilerFlags(CompilerFlags(0x10000), CompilerFlags(0x8000))
        assert result == CompilerFlags(0x18000)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_from_names_multi_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        result = CompilerFlags("with_statement", "print_function")
        assert result == CompilerFlags(0x18000)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_from_names_single_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        result = CompilerFlags("with_statement")
        assert result == CompilerFlags(0x8000)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_from_names_list_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        result = CompilerFlags(["with_statement", "print_function"])
        assert result == CompilerFlags(0x18000)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_from_mixed_multi_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        result = CompilerFlags("print_function", 0x8000, CompilerFlags("division"))
        assert result == CompilerFlags(0x1a000)


@skip_if_new_compiler_flags_values
def test_CompilerFlags_from_ast_1():
    node = ast.parse("from __future__ import with_statement, print_function")
    result = CompilerFlags(node)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert result == CompilerFlags(0x18000)


def test_CompilerFlags_bad_name_1():
    with pytest.raises(ValueError):
        CompilerFlags("print_statement")


def test_CompilerFlags_bad_int_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        with pytest.raises(ValueError):
            CompilerFlags(1)


def test_CompilerFlags_bad_multi_int_1():
    with pytest.raises(ValueError):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            CompilerFlags((0x8000 ^ 1), 0)




# To make tidy-imports automatically transform imports, create a dictionary
# like this:
#   __canonical_imports__ = {
#      "oldmodule.oldname": "newmodule.newname"
#   }
#
# You can write this in any *.py file in your $PYFLYBY_PATH,
# e.g. ~/.pyflyby/canonical.py.




from   Crypto.Cipher            import AES
import IPython
import blist
from   blist                    import sorteddict
import cssutils
import dateutil
import dateutil.parser
import kerberos
import mutagen
import perl
import pexpect
import pstats
import pyflyby
from   pyflyby                  import SaveframeReader, saveframe, xreload
import pylab
import pyodbc
import pysvn
import pytest
import pytz
import requests
import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.sql
import sympy
import xlrd
import yaml
from   yaml                     import MarkedYAMLError, YAMLError, YAMLObject



# pyflyby/_file.py.
# Copyright (C) 2011, 2012, 2013, 2014, 2015, 2018 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT
from __future__ import annotations

from   functools                import cached_property, total_ordering
import io
import os
import re
import sys
from   typing                   import ClassVar, List, Optional, Tuple, Union

from   pyflyby._util            import cmp, memoize

if sys.version_info < (3,10):
    NoneType = type(None)
else:
    from types import NoneType


class UnsafeFilenameError(ValueError):
    pass


# TODO: statcache

@total_ordering
class Filename(object):
    """
    A filename.

      >>> Filename('/etc/passwd')
      Filename('/etc/passwd')

    """
    _filename: str
    STDIN: Filename

    def __new__(cls, arg):
        if isinstance(arg, cls):
            # TODO make this assert False
            return cls._from_filename(arg._filename)
        if isinstance(arg, str):
            return cls._from_filename(arg)
        raise TypeError

    @classmethod
    def _from_filename(cls, filename: str):
        if not isinstance(filename, str):
            raise TypeError
        filename = os.path.abspath(filename)
        if not filename:
            raise UnsafeFilenameError("(empty string)")
        # we only allow filename with given character set
        match = re.search("[^a-zA-Z0-9_=+{}/.,~@-]", filename)
        if match:
            raise UnsafeFilenameError((filename, match))
        if re.search("(^|/)~", filename):
            raise UnsafeFilenameError(filename)
        self = object.__new__(cls)
        self._filename =  filename
        return self

    def __str__(self):
        return self._filename

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self._filename)

    def __truediv__(self, x):
        return type(self)(os.path.join(self._filename, x))

    def __hash__(self):
        return hash(self._filename)

    def __eq__(self, o):
        if self is o:
            return True
        if not isinstance(o, Filename):
            return NotImplemented
        return self._filename == o._filename

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, o):
        if not isinstance(o, Filename):
            return NotImplemented
        return self._filename < o._filename

    def __cmp__(self, o):
        if self is o:
            return 0
        if not isinstance(o, Filename):
            return NotImplemented
        return cmp(self._filename, o._filename)

    @cached_property
    def ext(self):
        """
        Returns the extension of this filename, including the dot.
        Returns ``None`` if no extension.

        :rtype:
          ``str`` or ``None``
        """
        lhs, dot, rhs = self._filename.rpartition('.')
        if not dot:
            return None
        return dot + rhs

    @cached_property
    def base(self):
        return os.path.basename(self._filename)

    @cached_property
    def dir(self):
        return type(self)(os.path.dirname(self._filename))

    @cached_property
    def real(self):
        return type(self)(os.path.realpath(self._filename))

    @property
    def realpath(self):
        return type(self)(os.path.realpath(self._filename))

    @property
    def exists(self):
        return os.path.exists(self._filename)

    @property
    def islink(self):
        return os.path.islink(self._filename)

    @property
    def isdir(self):
        return os.path.isdir(self._filename)

    @property
    def isfile(self):
        return os.path.isfile(self._filename)

    @property
    def isreadable(self):
        return os.access(self._filename, os.R_OK)

    @property
    def iswritable(self):
        return os.access(self._filename, os.W_OK)

    @property
    def isexecutable(self):
        return os.access(self._filename, os.X_OK)

    def startswith(self, prefix):
        prefix = Filename(prefix)
        if self == prefix:
            return True
        return self._filename.startswith("%s/" % (prefix,))

    def list(self, ignore_unsafe=True):
        filenames = [os.path.join(self._filename, f)
                     for f in sorted(os.listdir(self._filename))]
        result = []
        for f in filenames:
            try:
                f = Filename(f)
            except UnsafeFilenameError:
                if ignore_unsafe:
                    continue
                else:
                    raise
            result.append(f)
        return result

    @property
    def ancestors(self):
        """
        Return ancestors of self, from self to /.

          >>> Filename("/aa/bb").ancestors
          (Filename('/aa/bb'), Filename('/aa'), Filename('/'))

        :rtype:
          ``tuple`` of ``Filename`` s
        """
        result = [self]
        while True:
            dir = result[-1].dir
            if dir == result[-1]:
                break
            result.append(dir)
        return tuple(result)


@memoize
def _get_PATH():
    PATH = os.environ.get("PATH", "").split(os.pathsep)
    result = []
    for path in PATH:
        if not path:
            continue
        try:
            result.append(Filename(path))
        except UnsafeFilenameError:
            continue
    return tuple(result)


def which(program):
    """
    Find ``program`` on $PATH.

    :type program:
      ``str``
    :rtype:
      `Filename`
    :return:
      Program on $PATH, or ``None`` if not found.
    """
    # See if it exists in the current directory.
    candidate = Filename(program)
    if candidate.isreadable:
        return candidate
    for path in _get_PATH():
        candidate = path / program
        if candidate.isexecutable:
            return candidate
    return None



Filename.STDIN = Filename("/dev/stdin")

@total_ordering
class FilePos(object):
    """
    A (lineno, colno) position within a `FileText`.
    Both lineno and colno are 1-indexed.
    """

    lineno: int
    colno: int

    _ONE_ONE: ClassVar[FilePos]

    def __new__(cls, *args):
        if len(args) == 0:
            return cls._ONE_ONE
        if len(args) == 1:
            arg, = args
            if isinstance(arg, cls):
                return arg
            elif arg is None:
                return cls._ONE_ONE
            elif isinstance(arg, tuple):
                args = arg
                # Fall through
            else:
                raise TypeError
        lineno, colno = cls._intint(args)
        if lineno == colno == 1:
            return cls._ONE_ONE # space optimization
        if lineno < 1:
            raise ValueError(
                "FilePos: invalid lineno=%d; should be >= 1" % lineno,)
        if colno < 1:
            raise ValueError(
                "FilePos: invalid colno=%d; should be >= 1" % colno,)
        return cls._from_lc(lineno, colno)

    @staticmethod
    def _intint(args):
        if (type(args) is tuple and
            len(args) == 2 and
            type(args[0]) is type(args[1]) is int):
            return args
        else:
            raise TypeError("Expected (int,int); got %r" % (args,))

    @classmethod
    def _from_lc(cls, lineno:int, colno:int):
        self = object.__new__(cls)
        self.lineno = lineno
        self.colno  = colno
        return self

    def __add__(self, delta):
        '''
        "Add" a coordinate (line,col) delta to this ``FilePos``.

        Note that addition here may be a non-obvious.  If there is any line
        movement, then the existing column number is ignored, and the new
        column is the new column delta + 1 (to convert into 1-based numbers).

        :rtype:
          `FilePos`
        '''
        ldelta, cdelta = self._intint(delta)
        assert ldelta >= 0 and cdelta >= 0
        if ldelta == 0:
            return FilePos(self.lineno, self.colno + cdelta)
        else:
            return FilePos(self.lineno + ldelta, 1 + cdelta)

    def __str__(self):
        return "(%d,%d)" % (self.lineno, self.colno)

    def __repr__(self):
        return "FilePos%s" % (self,)

    @property
    def _data(self):
        return (self.lineno, self.colno)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, FilePos):
            return NotImplemented
        return self._data == other._data

    def __ne__(self, other):
        return not (self == other)

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, FilePos):
            return NotImplemented
        return cmp(self._data, other._data)

    # The rest are defined by total_ordering
    def __lt__(self, other):
        if self is other:
            return 0
        if not isinstance(other, FilePos):
            return NotImplemented
        return self._data < other._data

    def __hash__(self):
        return hash(self._data)



FilePos._ONE_ONE = FilePos._from_lc(1, 1)


@total_ordering
class FileText:
    """
    Represents a contiguous sequence of lines from a file.
    """

    filename: Optional[Filename]
    startpos: FilePos
    _lines: Optional[Tuple[str, ...]] = None

    def __new__(cls, arg, filename=None, startpos=None):
        """
        Return a new ``FileText`` instance.

        :type arg:
          ``FileText``, ``Filename``, ``str``, or tuple of ``str``
        :param arg:
          If a sequence of lines, then each should end with a newline and have
          no other newlines.  Otherwise, something that can be interpreted or
          converted into a sequence of lines.
        :type filename:
          `Filename`
        :param filename:
          Filename to attach to this ``FileText``, if not already given by
          ``arg``.
        :type startpos:
          ``FilePos``
        :param startpos:
          Starting file position (lineno & colno) of this ``FileText``, if not
          already given by ``arg``.
        :rtype:
          ``FileText``
        """
        if isinstance(filename, str):
            filename = Filename(filename)
        if isinstance(arg, cls):
            if filename is startpos is None:
                return arg
            return arg.alter(filename=filename, startpos=startpos)
        elif isinstance(arg, Filename):
            return cls(read_file(arg), filename=filename, startpos=startpos)
        elif hasattr(arg, "__text__"):
            return FileText(arg.__text__(), filename=filename, startpos=startpos)
        elif isinstance(arg, str):
            self = object.__new__(cls)
            self._lines = tuple(arg.split('\n'))
        else:
            raise TypeError("%s: unexpected %s"
                            % (cls.__name__, type(arg).__name__))

        assert isinstance(filename, (Filename, NoneType))
        startpos = FilePos(startpos)
        self.filename = filename
        self.startpos = startpos
        return self

    @classmethod
    def _from_lines(cls, lines, filename: Optional[Filename], startpos: FilePos):
        assert type(lines) is tuple
        assert len(lines) > 0
        assert isinstance(lines[0], str)
        assert not lines[-1].endswith("\n")
        assert isinstance(startpos, FilePos), repr(startpos)
        assert isinstance(filename, (Filename, type(None))), repr(filename)
        self = object.__new__(cls)
        self._lines    = tuple(lines)
        self.filename = filename
        self.startpos = startpos
        return self

    @cached_property
    def lines(self) -> Tuple[str, ...]:
        r"""
        Lines that have been split by newline.

        These strings do NOT contain '\n'.

        If the input file ended in '\n', then the last item will be the empty
        string.  This is to avoid having to check lines[-1].endswith('\n')
        everywhere.

        :rtype:
          ``tuple`` of ``str``
        """
        if self._lines is not None:
            return self._lines
        # Used if only initialized with 'joined'.
        # We use str.split() instead of str.splitlines() because the latter
        # doesn't distinguish between strings that end in newline or not
        # (or requires extra work to process if we use splitlines(True)).
        return tuple(self.joined.split('\n'))

    @cached_property
    def joined(self) -> str:
        return '\n'.join(self.lines)


    @classmethod
    def from_filename(cls, filename):
        return cls.from_lines(Filename(filename))

    def alter(self, filename: Optional[Filename] = None, startpos=None):
        if filename is not None:
            assert isinstance(filename, Filename)
        else:
            filename = self.filename
        if startpos is not None:
            startpos = FilePos(startpos)
        else:
            startpos = self.startpos
        if filename == self.filename and startpos == self.startpos:
            return self
        else:
            result = object.__new__(type(self))
            result._lines   = self._lines
            result.filename = filename
            result.startpos = startpos
            return result

    @cached_property
    def endpos(self):
        """
        The position after the last character in the text.

        :rtype:
          ``FilePos``
        """
        startpos = self.startpos
        lines    = self.lines
        lineno   = startpos.lineno + len(lines) - 1
        if len(lines) == 1:
            colno = startpos.colno + len(lines[-1])
        else:
            colno = 1 + len(lines[-1])
        return FilePos(lineno, colno)

    def _lineno_to_index(self, lineno):
        lineindex = lineno - self.startpos.lineno
        # Check that the lineindex is in range.  We don't allow pointing at
        # the line after the last line because we already ensured that
        # self.lines contains an extra empty string if necessary, to indicate
        # a trailing newline in the file.
        if not 0 <= lineindex < len(self.lines):
            raise IndexError(
                "Line number %d out of range [%d, %d)"
                % (lineno, self.startpos.lineno, self.endpos.lineno))
        return lineindex

    def _colno_to_index(self, lineindex, colno):
        coloffset = self.startpos.colno if lineindex == 0 else 1
        colindex = colno - coloffset
        line = self.lines[lineindex]
        # Check that the colindex is in range.  We do allow pointing at the
        # character after the last (non-newline) character in the line.
        if not 0 <= colindex <= len(line):
            raise IndexError(
                "Column number %d on line %d out of range [%d, %d]"
                % (colno, lineindex+self.startpos.lineno,
                   coloffset, coloffset+len(line)))
        return colindex

    def __getitem__(self, arg):
        """
        Return the line(s) with the given line number(s).
        If slicing, returns an instance of ``FileText``.

        Note that line numbers are indexed based on ``self.startpos.lineno``
        (which is 1 at the start of the file).

          >>> FileText("a\\nb\\nc\\nd")[2]
          'b'

          >>> FileText("a\\nb\\nc\\nd")[2:4]
          FileText('b\\nc\\n', startpos=(2,1))

          >>> FileText("a\\nb\\nc\\nd")[0]
          Traceback (most recent call last):
          ...
          IndexError: Line number 0 out of range [1, 4)

        When slicing, the input arguments can also be given as ``FilePos``
        arguments or (lineno,colno) tuples.  These are 1-indexed at the start
        of the file.

          >>> FileText("a\\nb\\nc\\nd")[(2,2):4]
          FileText('\\nc\\n', startpos=(2,2))

        :rtype:
          ``str`` or `FileText`
        """
        L = self._lineno_to_index
        C = self._colno_to_index
        if isinstance(arg, slice):
            if arg.step is not None and arg.step != 1:
                raise ValueError("steps not supported")
            # Interpret start (lineno,colno) into indexes.
            if arg.start is None:
                start_lineindex = 0
                start_colindex = 0
            elif isinstance(arg.start, int):
                start_lineindex = L(arg.start)
                start_colindex = 0
            else:
                startpos = FilePos(arg.start)
                start_lineindex = L(startpos.lineno)
                start_colindex = C(start_lineindex, startpos.colno)
            # Interpret stop (lineno,colno) into indexes.
            if arg.stop is None:
                stop_lineindex = len(self.lines)
                stop_colindex = len(self.lines[-1])
            elif isinstance(arg.stop, int):
                stop_lineindex = L(arg.stop)
                stop_colindex = 0
            else:
                stoppos = FilePos(arg.stop)
                stop_lineindex = L(stoppos.lineno)
                stop_colindex = C(stop_lineindex, stoppos.colno)
            # {start,stop}_{lineindex,colindex} are now 0-indexed
            # [open,closed) ranges.
            assert 0 <= start_lineindex <= stop_lineindex < len(self.lines)
            assert 0 <= start_colindex <= len(self.lines[start_lineindex])
            assert 0 <= stop_colindex <= len(self.lines[stop_lineindex])
            # Optimization: return entire range
            if (start_lineindex == 0 and
                start_colindex == 0 and
                stop_lineindex == len(self.lines)-1 and
                stop_colindex == len(self.lines[-1])):
                return self
            # Get the lines we care about.  We always include an extra entry
            # at the end which we'll chop to the desired number of characters.
            result_split = list(self.lines[start_lineindex:stop_lineindex+1])
            # Clip the starting and ending strings.  We do the end clip first
            # in case the result has only one line.
            result_split[-1] = result_split[-1][:stop_colindex]
            result_split[0] = result_split[0][start_colindex:]
            # Compute the new starting line and column numbers.
            result_lineno = start_lineindex + self.startpos.lineno
            if start_lineindex == 0:
                result_colno = start_colindex + self.startpos.colno
            else:
                result_colno = start_colindex + 1
            result_startpos = FilePos(result_lineno, result_colno)
            return FileText._from_lines(tuple(result_split),
                                        filename=self.filename,
                                        startpos=result_startpos)
        elif isinstance(arg, int):
            # Return a single line.
            lineindex = L(arg)
            return self.lines[lineindex]
        else:
            raise TypeError("bad type %r" % (type(arg),))

    @classmethod
    def concatenate(cls, args):
        """
        Concatenate a bunch of `FileText` arguments.  Uses the ``filename``
        and ``startpos`` from the first argument.

        :rtype:
          `FileText`
        """
        args = [FileText(x) for x in args]
        if len(args) == 1:
            return args[0]
        return FileText(
            ''.join([l.joined for l in args]),
            filename=args[0].filename,
            startpos=args[0].startpos)

    def __repr__(self):
        r = "%s(%r" % (type(self).__name__, self.joined,)
        if self.filename is not None:
            r += ", filename=%r" % (str(self.filename),)
        if self.startpos != FilePos():
            r += ", startpos=%s" % (self.startpos,)
        r += ")"
        return r

    def __str__(self):
        return self.joined

    def __eq__(self, o):
        if self is o:
            return True
        if not isinstance(o, FileText):
            return NotImplemented
        return (self.filename == o.filename and
                self.joined   == o.joined   and
                self.startpos == o.startpos)

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, o):
        if not isinstance(o, FileText):
            return NotImplemented
        return ((self.filename, self.joined, self.startpos) <
                   (o   .filename, o   .joined, o   .startpos))

    def __cmp__(self, o):
        if self is o:
            return 0
        if not isinstance(o, FileText):
            return NotImplemented
        return cmp((self.filename, self.joined, self.startpos),
                   (o   .filename, o   .joined, o   .startpos))

    def __hash__(self):
        h = hash((self.filename, self.joined, self.startpos))
        self.__hash__ = lambda: h
        return h


def read_file(filename: Filename) -> FileText:
    assert isinstance(filename, Filename)
    if filename == Filename.STDIN:
        data = sys.stdin.read()
    else:
        with io.open(str(filename), 'r') as f:
            data = f.read()
    return FileText(data, filename=filename)


def write_file(filename: Filename, data):
    assert isinstance(filename, Filename)
    data = FileText(data)
    with open(str(filename), 'w') as f:
        f.write(data.joined)

def atomic_write_file(filename: Filename, data):
    assert isinstance(filename, Filename)
    data = FileText(data)
    temp_filename = Filename("%s.tmp.%s" % (filename, os.getpid(),))
    write_file(temp_filename, data)
    try:
        st = os.stat(str(filename)) # OSError if file didn't exit before
        os.chmod(str(temp_filename), st.st_mode)
        os.chown(str(temp_filename), -1, st.st_gid) # OSError if not member of group
    except OSError:
        pass
    os.rename(str(temp_filename), str(filename))


def expand_py_files_from_args(
    pathnames: Union[List[Filename], Filename], on_error=lambda filename: None
):
    """
    Enumerate ``*.py`` files, recursively.

    Arguments that are files are always included.
    Arguments that are directories are recursively searched for ``*.py`` files.

    :type pathnames:
      ``list`` of `Filename` s
    :type on_error:
      callable
    :param on_error:
      Function that is called for arguments directly specified in ``pathnames``
      that don't exist or are otherwise inaccessible.
    :rtype:
      ``list`` of `Filename` s
    """
    if not isinstance(pathnames, (tuple, list)):
        # July 2024 DeprecationWarning
        # this seem to be used only internally, maybe deprecate not passing a list.
        pathnames = [pathnames]
    for f in pathnames:
        assert isinstance(f, Filename)
    result = []
    # Check for problematic arguments.  Note that we intentionally only do
    # this for directly specified arguments, not for recursively traversed
    # arguments.
    stack = []
    for pathname in reversed(pathnames):
        if pathname.isfile:
            stack.append((pathname, True))
        elif pathname.isdir:
            stack.append((pathname, False))
        else:
            on_error(pathname)
    while stack:
        pathname, isfile = stack.pop(-1)
        if isfile:
            result.append(pathname)
            continue
        for f in reversed(pathname.list()):
            # Check inclusions/exclusions for recursion.  Note that we
            # intentionally do this in the recursive step rather than the
            # base step because if the user specification includes
            # e.g. .pyflyby, we do want to include it; however, we don't
            # want to recurse into .pyflyby ourselves.
            if f.base.startswith("."):
                continue
            if f.base == "__pycache__":
                continue
            if f.isfile:
                if f.ext == ".py":
                    stack.append((f, True))
            elif f.isdir:
                stack.append((f, False))
            else:
                # Silently ignore non-files/dirs from traversal.
                pass
    return result




# Some self tests to check that our test setup is pointing to wrong paths etc.



import os
import pytest
import subprocess
import sys

import pyflyby


def pipe(command, stdin=""):
    return subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT
    ).communicate(stdin)[0].strip().decode('utf-8')


def test_pytest_version_1():
    pytest_version = tuple(map(int, pytest.__version__.split(".")[:2]))
    if pytest_version < (2, 4):
        raise AssertionError("test cases require pytest >= 2.4; "
                             "your version is pytest %s", pytest.__version__)


def test_pyflyby_version_1():
    # Check that the version that we've imported here is the same as the one
    # in this repository.
    PYFLYBY_HOME   = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
    PYFLYBY_PYPATH = os.path.join(PYFLYBY_HOME, "lib/python")
    version_vars = {}
    version_fn = os.path.join(PYFLYBY_PYPATH, "pyflyby/_version.py")
    exec(open(version_fn).read(), {}, version_vars)
    expected = version_vars["__version__"]
    result = pyflyby.__version__
    assert expected == result


def test_pyflyby_tox_path_1():
    # If we're inside tox, then check that we've loaded the virtualenv version.
    if '/.tox/' not in sys.prefix:
        return
    assert pyflyby.__file__.startswith(os.path.normpath(sys.prefix))


def test_pyflyby_subprocess_file_1():
    # Check that our test setup is getting the right pyflyby.
    cmd = "import os, pyflyby; print(os.path.realpath(pyflyby.__file__))"
    result = pipe([sys.executable, '-c', cmd]).replace(".pyc", ".py")
    expected = os.path.realpath(pyflyby.__file__).replace(".pyc", ".py")
    assert expected == result


def test_pyflyby_subprocess_version_1():
    # Check that our test setup is getting the right pyflyby.
    cmd = "import pyflyby; print(pyflyby.__version__)"
    result = pipe([sys.executable, '-c', cmd])
    expected = pyflyby.__version__
    assert expected == result



"""
Virtual module to create dynamic import at runtime.

It is sometime desirable to have auto import which are define only during
a session and never exist on a on-disk file.

This is injects a Dict module loader as well as a dictionary registry of in
memory module.

This is mostly use in IPython for lazy variable initialisation without having
to use proxy objects.

To use, put the following in your IPython startup files
(``~/.ipython/profile_default/startup/autoimp.py`), or in your IPython
configuration file:


.. code:: python

    from pyflyby._dynimp import add_import

    add_import("foo", "foo = 1")

    add_import(
        "df, data",
        '''
        import pandas as pd
        data = [1,2,3]
        df =  pd.DataFrame(data)
    ''',
    )

Now at the IPython prompt, if the pyflyby extension is loaded (either because
you started  using the ``py`` cli, or some configuration options like  ``ipython
--TerminalIPythonApp.extra_extensions=pyflyby``. When trying to use an undefined
variable like ``foo``, ``df`` or ``data``, the corresponding module will be
executed and the relevant variable imported.


"""
import importlib.abc
import importlib.util
import sys

from textwrap import dedent
from typing import FrozenSet

from pyflyby._importclns import ImportSet, Import

module_dict = {}

PYFLYBY_LAZY_LOAD_PREFIX = "from pyflyby_autoimport_"

def add_import(names: str, code: str, *, strict: bool = True):
    """
    Add a runtime generated import module

    Parameters
    ----------
    names: str
        name, or comma separated list variable names that should be created by
        executing and importing `code`.
    code: str
        potentially multiline string that will be turned into a module,
        executed and from which variables listed in names can be imported.
    strict: bool
        Raise in case of problem loading IPython of if pyflyby extension not installed.
        otherwise just ignore error



    Examples
    --------

    >>> add_import('pd, df', '''
    ...     import pandas a pd
    ...
    ...     df = pd.DataFrame([[1,2], [3,4]])
    ... ''', strict=False)  # don't fail doctest

    """
    try:
        ip = _raise_if_problem()
    except Exception:
        if strict:
            raise
        else:
            return
    return _add_import(ip, names, code)


def _raise_if_problem():
    try:
        import IPython
    except ModuleNotFoundError as e:
        raise ImportError("Dynamic autoimport requires IPython to be installed") from e

    ip = IPython.get_ipython()
    if ip is None:
        raise ImportError("Dynamic autoimport only work from within IPython")

    if not hasattr(ip, "_auto_importer"):
        raise ValueError(
            "IPython needs to be loaded with pyflyby extension for lazy variable to work"
        )
    return ip


def _add_import(ip, names: str, code: str) -> None:
    """
    private version of add_import
    """
    assert ip is not None
    module = PYFLYBY_LAZY_LOAD_PREFIX.split()[1]
    mang = module + names.replace(",", "_").replace(" ", "_")
    a: FrozenSet[Import] = ImportSet(f"from {mang} import {names}")._importset
    b: FrozenSet[Import] = ip._auto_importer.db.known_imports._importset
    s_import: FrozenSet[Import] = a | b

    ip._auto_importer.db.known_imports = ImportSet._from_imports(list(s_import))
    module_dict[mang] = dedent(code)

class DictLoader(importlib.abc.Loader):
    """
    A dict based loader for in-memory module definition.
    """
    def __init__(self, module_name, module_code):
        self.module_name = module_name
        self.module_code = module_code

    def create_module(self, spec):
        return None  # Use default module creation semantics

    def exec_module(self, module):
        """
        we exec module code directly in memory
        """
        exec(self.module_code, module.__dict__)


class DictFinder(importlib.abc.MetaPathFinder):
    """
    A meta path finder for abode DictLoader
    """
    def find_spec(self, fullname, path, target=None):
        if fullname in module_dict:
            module_code = module_dict[fullname]
            loader = DictLoader(fullname, module_code)
            return importlib.util.spec_from_loader(fullname, loader)
        return None


def inject():
    sys.meta_path.insert(0, DictFinder())



# pyflyby/_imports2s.py.
# Copyright (C) 2011-2018 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT

from   pyflyby._autoimp         import scan_for_import_issues
from   pyflyby._file            import FileText, Filename
from   pyflyby._flags           import CompilerFlags
from   pyflyby._importclns      import ImportSet, NoSuchImportError
from   pyflyby._importdb        import ImportDB
from   pyflyby._importstmt      import ImportFormatParams, ImportStatement
from   pyflyby._log             import logger
from   pyflyby._parse           import PythonBlock
from   pyflyby._util            import ImportPathCtx, Inf, NullCtx, memoize
import re

from typing import Union, Optional, Literal

from textwrap import indent


class SourceToSourceTransformationBase:

    input: PythonBlock

    def __new__(cls, arg):
        if isinstance(arg, cls):
            return arg
        if isinstance(arg, (PythonBlock, FileText, Filename, str)):
            return cls._from_source_code(arg)
        raise TypeError("%s: got unexpected %s"
                        % (cls.__name__, type(arg).__name__))

    @classmethod
    def _from_source_code(cls, codeblock):
        # TODO: don't do that.
        self = object.__new__(cls)
        if isinstance(codeblock, PythonBlock):
            self.input = codeblock
        elif isinstance(codeblock, FileText):
            self.input = PythonBlock(codeblock)
        else:
            if not codeblock.endswith('\n'):
                codeblock += '\n'
            self.input = PythonBlock(codeblock)
        self.preprocess()
        return self

    def preprocess(self):
        pass

    def pretty_print(self, params=None):
        raise NotImplementedError

    def output(self, params=None) -> PythonBlock:
        """
        Pretty-print and return as a `PythonBlock`.

        :rtype:
          `PythonBlock`
        """
        result = self.pretty_print(params=params)
        result = PythonBlock(result, filename=self.input.filename)
        return result

    def __repr__(self):
        return f"<{self.__class__.__name__}\n{indent(str(self.pretty_print()),'    ')}\n at 0x{hex(id(self))}>"


class SourceToSourceTransformation(SourceToSourceTransformationBase):

    _output: PythonBlock

    def preprocess(self):
        assert isinstance(self.input, PythonBlock), self.input
        self._output = self.input

    def pretty_print(self, params=None):
        return self._output.text


class SourceToSourceImportBlockTransformation(SourceToSourceTransformationBase):
    def preprocess(self):
        self.importset = ImportSet(self.input, ignore_shadowed=True)

    def pretty_print(self, params=None):
        params = ImportFormatParams(params)
        return self.importset.pretty_print(params)

    def __repr__(self):
        return f"<SourceToSourceImportBlockTransformation {self.importset!r} @{hex(id(self))}>"


class LineNumberNotFoundError(Exception):
    pass

class LineNumberAmbiguousError(Exception):
    pass

class NoImportBlockError(Exception):
    pass

class ImportAlreadyExistsError(Exception):
    pass

class SourceToSourceFileImportsTransformation(SourceToSourceTransformationBase):
    def preprocess(self):
        # Group into blocks of imports and non-imports.  Get a sequence of all
        # imports for the transformers to operate on.
        self.blocks = []
        self.import_blocks = []
        for is_imports, subblock in self.input.groupby(lambda ps: ps.is_import):
            if is_imports:
                trans = SourceToSourceImportBlockTransformation(subblock)
                self.import_blocks.append(trans)
            else:
                trans = SourceToSourceTransformation(subblock)
            self.blocks.append(trans)

    def pretty_print(self, params=None):
        params = ImportFormatParams(params)
        result = [block.pretty_print(params=params) for block in self.blocks]
        return FileText.concatenate(result)

    def find_import_block_by_lineno(self, lineno: int):
        """
        Find the import block containing the given line number.

        :type lineno:
          ``int``
        :rtype:
          `SourceToSourceImportBlockTransformation`
        """
        results = [
            b
            for b in self.import_blocks
            if b.input.startpos.lineno <= lineno <= b.input.endpos.lineno]
        if len(results) == 0:
            raise LineNumberNotFoundError(lineno)
        if len(results) > 1:
            raise LineNumberAmbiguousError(lineno)
        return results[0]

    def remove_import(self, imp, lineno):
        """
        Remove the given import.

        :type imp:
          `Import`
        :type lineno:
          ``int``
        """
        block = self.find_import_block_by_lineno(lineno)
        try:
            imports = block.importset.by_import_as[imp.import_as]
        except KeyError:
            raise NoSuchImportError
        assert len(imports)
        if len(imports) > 1:
            raise Exception("Multiple imports to remove: %r" % (imports,))
        imp = imports[0]
        block.importset = block.importset.without_imports([imp])
        return imp

    def select_import_block_by_closest_prefix_match(self, imp, max_lineno):
        """
        Heuristically pick an import block that ``imp`` "fits" best into.  The
        selection is based on the block that contains the import with the
        longest common prefix.

        :type imp:
          `Import`
        :param max_lineno:
          Only return import blocks earlier than ``max_lineno``.
        :rtype:
          `SourceToSourceImportBlockTransformation`
        """
        # Create a data structure that annotates blocks with data by which
        # we'll sort.
        annotated_blocks = [
            ( (max([0] + [len(imp.prefix_match(oimp))
                          for oimp in block.importset.imports]),
               block.input.endpos.lineno),
              block )
            for block in self.import_blocks
            if block.input.endpos.lineno <= max_lineno+1 ]
        if not annotated_blocks:
            raise NoImportBlockError()
        annotated_blocks.sort()
        if imp.split.module_name == '__future__':
            # For __future__ imports, only add to an existing block that
            # already contains __future__ import(s).  If there are no existing
            # import blocks containing __future__, don't return any result
            # here, so that we will add a new one at the top.
            if not annotated_blocks[-1][0][0] > 0:
                raise NoImportBlockError
        return annotated_blocks[-1][1]

    def insert_new_blocks_after_comments(self, blocks):
        blocks = [SourceToSourceTransformationBase(block) for block in blocks]
        if isinstance(self.blocks[0], SourceToSourceImportBlockTransformation):
            # Kludge.  We should add an "output" attribute to
            # SourceToSourceImportBlockTransformation and enumerate over that,
            # instead of enumerating over the input below.
            self.blocks[0:0] = blocks
            return
        # Get the "statements" in the first block.
        statements = self.blocks[0].input.statements
        # Find the insertion point.
        for idx, statement in enumerate(statements):
            if not statement.is_comment_or_blank_or_string_literal:
                if idx == 0:
                    # First block starts with a noncomment, so insert before
                    # it.
                    self.blocks[0:0] = blocks
                else:
                    # Found a non-comment after comment, so break it up and
                    # insert in the middle.
                    self.blocks[:1] = (
                        [SourceToSourceTransformation(
                            PythonBlock.concatenate(statements[:idx]))] +
                        blocks +
                        [SourceToSourceTransformation(
                            PythonBlock.concatenate(statements[idx:]))])
                break
        else:
            # First block is entirely comments, so just insert after it.
            self.blocks[1:1] = blocks

    def insert_new_import_block(self):
        """
        Adds a new empty imports block.  It is added before the first
        non-comment statement.  Intended to be used when the input contains no
        import blocks (before uses).
        """
        block = SourceToSourceImportBlockTransformation("")
        sepblock = SourceToSourceTransformation("")
        sepblock._output = PythonBlock("\n")
        self.insert_new_blocks_after_comments([block, sepblock])
        self.import_blocks.insert(0, block)
        return block

    def add_import(self, imp, lineno=Inf):
        """
        Add the specified import.  Picks an existing global import block to
        add to, or if none found, creates a new one near the beginning of the
        module.

        :type imp:
          `Import`
        :param lineno:
          Line before which to add the import.  ``Inf`` means no constraint.
        """
        try:
            block = self.select_import_block_by_closest_prefix_match(
                imp, lineno)
        except NoImportBlockError:
            block = self.insert_new_import_block()
        if imp in block.importset.imports:
            raise ImportAlreadyExistsError(imp)
        block.importset = block.importset.with_imports([imp])


def reformat_import_statements(codeblock, params=None):
    r"""
    Reformat each top-level block of import statements within a block of code.
    Blank lines, comments, etc. are left alone and separate blocks of imports.

    Parse the entire code block into an ast, group into consecutive import
    statements and other lines.  Each import block consists entirely of
    'import' (or 'from ... import') statements.  Other lines, including blanks
    and comment lines, are not touched.

      >>> print(reformat_import_statements(
      ...     'from foo import bar2 as bar2x, bar1\n'
      ...     'import foo.bar3 as bar3x\n'
      ...     'import foo.bar4\n'
      ...     '\n'
      ...     'import foo.bar0 as bar0\n').text.joined)
      import foo.bar4
      from foo import bar1, bar2 as bar2x, bar3 as bar3x
      <BLANKLINE>
      from foo import bar0
      <BLANKLINE>

    :type codeblock:
      `PythonBlock` or convertible (``str``)
    :type params:
      `ImportFormatParams`
    :rtype:
      `PythonBlock`
    """
    params = ImportFormatParams(params)
    transformer = SourceToSourceFileImportsTransformation(codeblock)
    return transformer.output(params=params)


def ImportPathForRelativeImportsCtx(codeblock):
    """
    Context manager that temporarily modifies ``sys.path`` so that relative
    imports for the given ``codeblock`` work as expected.

    :type codeblock:
      `PythonBlock`
    """
    if not isinstance(codeblock, PythonBlock):
        codeblock = PythonBlock(codeblock)
    if not codeblock.filename:
        return NullCtx()
    if codeblock.flags & CompilerFlags("absolute_import"):
        return NullCtx()
    return ImportPathCtx(str(codeblock.filename.dir))


def fix_unused_and_missing_imports(
    codeblock: Union[PythonBlock, str, Filename],
    add_missing: bool = True,
    remove_unused: Union[Literal["AUTOMATIC"], bool] = "AUTOMATIC",
    add_mandatory: bool = True,
    db: Optional[ImportDB] = None,
    params=None,
) -> PythonBlock:
    r"""
    Check for unused and missing imports, and fix them automatically.

    Also formats imports.

    In the example below, ``m1`` and ``m3`` are unused, so are automatically
    removed.  ``np`` was undefined, so an ``import numpy as np`` was
    automatically added.

      >>> codeblock = PythonBlock(
      ...     'from foo import m1, m2, m3, m4\n'
      ...     'm2, m4, np.foo', filename="/tmp/foo.py")

      >>> print(fix_unused_and_missing_imports(codeblock, add_mandatory=False))
      [PYFLYBY] /tmp/foo.py: removed unused 'from foo import m1'
      [PYFLYBY] /tmp/foo.py: removed unused 'from foo import m3'
      [PYFLYBY] /tmp/foo.py: added 'import numpy as np'
      import numpy as np
      from foo import m2, m4
      m2, m4, np.foo

    :type codeblock:
      `PythonBlock` or convertible (``str``)
    :rtype:
      `PythonBlock`
    """
    _codeblock: PythonBlock
    if isinstance(codeblock, Filename):
        _codeblock = PythonBlock(codeblock)
    if not isinstance(codeblock, PythonBlock):
        _codeblock = PythonBlock(codeblock)
    else:
        _codeblock = codeblock
    if remove_unused == "AUTOMATIC":
        fn = _codeblock.filename
        remove_unused = not (fn and
                             (fn.base == "__init__.py"
                              or ".pyflyby" in str(fn).split("/")))
    elif remove_unused is True or remove_unused is False:
        pass
    else:
        raise ValueError("Invalid remove_unused=%r" % (remove_unused,))
    params = ImportFormatParams(params)
    db = ImportDB.interpret_arg(db, target_filename=_codeblock.filename)
    # Do a first pass reformatting the imports to get rid of repeated or
    # shadowed imports, e.g. L1 here:
    #   import foo  # L1
    #   import foo  # L2
    #   foo         # L3
    _codeblock = reformat_import_statements(_codeblock, params=params)

    filename = _codeblock.filename
    transformer = SourceToSourceFileImportsTransformation(_codeblock)
    missing_imports, unused_imports = scan_for_import_issues(
        _codeblock, find_unused_imports=remove_unused, parse_docstrings=True
    )
    logger.debug("missing_imports = %r", missing_imports)
    logger.debug("unused_imports = %r", unused_imports)
    if remove_unused and unused_imports:
        # Go through imports to remove.  [This used to be organized by going
        # through import blocks and removing all relevant blocks from there,
        # but if one removal caused problems the whole thing would fail.  The
        # CPU cost of calling without_imports() multiple times isn't worth
        # that.]
        # TODO: don't remove unused mandatory imports.  [This isn't
        # implemented yet because this isn't necessary for __future__ imports
        # since they aren't reported as unused, and those are the only ones we
        # have by default right now.]
        for lineno, imp in unused_imports:
            try:
                imp = transformer.remove_import(imp, lineno)
            except NoSuchImportError:
                logger.error(
                    "%s: couldn't remove import %r", filename, imp,)
            except LineNumberNotFoundError as e:
                logger.debug(
                    "%s: unused import %r on line %d not global",
                    filename, str(imp), e.args[0])
            else:
                logger.info("%s: removed unused '%s'", filename, imp)

    if add_missing and missing_imports:
        missing_imports.sort(key=lambda k: (k[1], k[0]))
        known = db.known_imports.by_import_as
        # Decide on where to put each import to be added.  Find the import
        # block with the longest common prefix.  Tie-break by preferring later
        # blocks.
        added_imports = set()
        for lineno, ident in missing_imports:
            import_as = ident.parts[0]
            try:
                imports = known[import_as]
            except KeyError:
                logger.warning(
                    "%s:%s: undefined name %r and no known import for it",
                    filename, lineno, import_as)
                continue
            if len(imports) != 1:
                logger.error("%s: don't know which of %r to use",
                             filename, imports)
                continue
            imp_to_add = imports[0]
            if imp_to_add in added_imports:
                continue
            transformer.add_import(imp_to_add, lineno)
            added_imports.add(imp_to_add)
            logger.info("%s: added %r", filename,
                        imp_to_add.pretty_print().strip())

    if add_mandatory:
        # Todo: allow not adding to empty __init__ files?
        mandatory = db.mandatory_imports.imports
        for imp in mandatory:
            try:
                transformer.add_import(imp)
            except ImportAlreadyExistsError:
                pass
            else:
                logger.info("%s: added mandatory %r",
                            filename, imp.pretty_print().strip())

    return transformer.output(params=params)


def remove_broken_imports(codeblock, params=None):
    """
    Try to execute each import, and remove the ones that don't work.

    Also formats imports.

    :type codeblock:
      `PythonBlock` or convertible (``str``)
    :rtype:
      `PythonBlock`
    """
    if not isinstance(codeblock, PythonBlock):
        codeblock = PythonBlock(codeblock)
    params = ImportFormatParams(params)
    filename = codeblock.filename
    transformer = SourceToSourceFileImportsTransformation(codeblock)
    for block in transformer.import_blocks:
        broken = []
        for imp in list(block.importset.imports):
            ns = {}
            try:
                exec(imp.pretty_print(), ns)
            except Exception as e:
                logger.info("%s: Could not import %r; removing it: %s: %s",
                            filename, imp.fullname, type(e).__name__, e)
                broken.append(imp)
        block.importset = block.importset.without_imports(broken)
    return transformer.output(params=params)


def replace_star_imports(codeblock, params=None):
    r"""
    Replace lines such as::

      from foo.bar import *
    with
      from foo.bar import f1, f2, f3

    Note that this requires involves actually importing ``foo.bar``, which may
    have side effects.  (TODO: rewrite to avoid this?)

    The result includes all imports from the ``email`` module.  The result
    excludes shadowed imports.  In this example:

        1. The original ``MIMEAudio`` import is shadowed, so it is removed.
        2. The ``MIMEImage`` import in the ``email`` module is shadowed by a
           subsequent import, so it is omitted.

        >>> codeblock = PythonBlock('from keyword import *', filename="/tmp/x.py")

        >>> print(replace_star_imports(codeblock)) # doctest: +SKIP
        [PYFLYBY] /tmp/x.py: replaced 'from keyword import *' with 2 imports
        from keyword import iskeyword, kwlist
        <BLANKLINE>

    Usually you'll want to remove unused imports after replacing star imports.

    :type codeblock:
      `PythonBlock` or convertible (``str``)
    :rtype:
      `PythonBlock`
    """
    from pyflyby._modules import ModuleHandle
    params = ImportFormatParams(params)
    if not isinstance(codeblock, PythonBlock):
        codeblock = PythonBlock(codeblock)
    filename = codeblock.filename
    transformer = SourceToSourceFileImportsTransformation(codeblock)
    for block in transformer.import_blocks:
        # Iterate over the import statements in ``block.input``.  We do this
        # instead of using ``block.importset`` because the latter doesn't
        # preserve the order of inputs.  The order is important for
        # determining what's shadowed.
        imports = [
            imp
            for s in block.input.statements
            for imp in ImportStatement(s).imports
        ]
        # Process "from ... import *" statements.
        new_imports = []
        for imp in imports:
            if imp.split.member_name != "*":
                new_imports.append(imp)
            elif imp.split.module_name.startswith("."):
                # The source contains e.g. "from .foo import *".  Right now we
                # don't have a good way to figure out the absolute module
                # name, so we can't get at foo.  That said, there's a decent
                # chance that this is inside an __init__ anyway, which is one
                # of the few justifiable use cases for star imports in library
                # code.
                logger.warning("%s: can't replace star imports in relative import: %s",
                               filename, imp.pretty_print().strip())
                new_imports.append(imp)
            else:
                module = ModuleHandle(imp.split.module_name)
                try:
                    with ImportPathForRelativeImportsCtx(codeblock):
                        exports = module.exports
                except Exception as e:
                    logger.warning(
                        "%s: couldn't import '%s' to enumerate exports, "
                        "leaving unchanged: '%s'.  %s: %s",
                        filename, module.name, imp, type(e).__name__, e)
                    new_imports.append(imp)
                    continue
                if not exports:
                    # We found nothing in the target module.  This probably
                    # means that module itself is just importing things from
                    # other modules.  Currently we intentionally exclude those
                    # imports since usually we don't want them.  TODO: do
                    # something better here.
                    logger.warning("%s: found nothing to import from %s, "
                                   "leaving unchanged: '%s'",
                                   filename, module, imp)
                    new_imports.append(imp)
                else:
                    new_imports.extend(exports)
                    logger.info("%s: replaced %r with %d imports", filename,
                                imp.pretty_print().strip(), len(exports))
        block.importset = ImportSet(new_imports, ignore_shadowed=True)
    return transformer.output(params=params)


def transform_imports(codeblock, transformations, params=None):
    """
    Transform imports as specified by ``transformations``.

    transform_imports() perfectly replaces all imports in top-level import
    blocks.

    For the rest of the code body, transform_imports() does a crude textual
    string replacement.  This is imperfect but handles most cases.  There may
    be some false positives, but this is difficult to avoid.  Generally we do
    want to do replacements even within in strings and comments.

      >>> result = transform_imports("from m import x", {"m.x": "m.y.z"})
      >>> print(result.text.joined.strip())
      from m.y import z as x

    :type codeblock:
      `PythonBlock` or convertible (``str``)
    :type transformations:
      ``dict`` from ``str`` to ``str``
    :param transformations:
      A map of import prefixes to replace, e.g. {"aa.bb": "xx.yy"}
    :rtype:
      `PythonBlock`
    """
    if not isinstance(codeblock, PythonBlock):
        codeblock = PythonBlock(codeblock)
    params = ImportFormatParams(params)
    transformer = SourceToSourceFileImportsTransformation(codeblock)
    @memoize
    def transform_import(imp):
        # Transform a block of imports.
        # TODO: optimize
        # TODO: handle transformations containing both a.b=>x and a.b.c=>y
        for k, v in transformations.items():
            imp = imp.replace(k, v)
        return imp
    def transform_block(block):
        # Do a crude string replacement in the PythonBlock.
        if not isinstance(block, PythonBlock):
            block = PythonBlock(block)
        s = block.text.joined
        for k, v in transformations.items():
            s = re.sub("\\b%s\\b" % (re.escape(k)), v, s)
        return PythonBlock(s, flags=block.flags)
    # Loop over transformer blocks.
    for block in transformer.blocks:
        if isinstance(block, SourceToSourceImportBlockTransformation):
            input_imports = block.importset.imports
            output_imports = [ transform_import(imp) for imp in input_imports ]
            block.importset = ImportSet(output_imports, ignore_shadowed=True)
        else:
            block._output = transform_block(block.input)
    return transformer.output(params=params)


def canonicalize_imports(codeblock, params=None, db=None):
    """
    Transform ``codeblock`` as specified by ``__canonical_imports__`` in the
    global import library.

    :type codeblock:
      `PythonBlock` or convertible (``str``)
    :rtype:
      `PythonBlock`
    """
    if not isinstance(codeblock, PythonBlock):
        codeblock = PythonBlock(codeblock)
    params = ImportFormatParams(params)
    db = ImportDB.interpret_arg(db, target_filename=codeblock.filename)
    transformations = db.canonical_imports
    return transform_imports(codeblock, transformations, params=params)



# pyflyby/_format.py.
# Copyright (C) 2011, 2012, 2013, 2014 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT



class FormatParams(object):
    max_line_length = None
    _max_line_lenght_default = 79
    wrap_paren = True
    indent = 4
    hanging_indent = 'never'
    use_black = False

    def __new__(cls, *args, **kwargs):
        if not kwargs and len(args) == 1 and isinstance(args[0], cls):
            return args[0]
        self = object.__new__(cls)
        # TODO: be more careful here
        dicts = []
        for arg in args:
            if arg is None:
                pass
            elif isinstance(arg, cls) or hasattr(self, "__dict__"):
                dicts.append(arg.__dict__)
            else:
                raise TypeError(
                    "expected None, or instance of %s cls, got %s" % (cls, arg)
                )
        if kwargs:
            dicts.append(kwargs)
        for kwargs in dicts:
            for key, value in kwargs.items():
                if hasattr(self, key):
                    setattr(self, key, value)
                else:
                    raise ValueError("bad kwarg %r" % (key,))
        return self

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.__dict__}>'


def fill(tokens, sep=(", ", ""), prefix="", suffix="", newline="\n",
         max_line_length=80):
    r"""
    Given a sequences of strings, fill them into a single string with up to
    ``max_line_length`` characters each.

      >>> fill(["'hello world'", "'hello two'"],
      ...            prefix=("print ", "      "), suffix=(" \\", ""),
      ...            max_line_length=25)
      "print 'hello world', \\\n      'hello two'\n"

    :param tokens:
      Sequence of strings to fill.  There must be at least one token.
    :param sep:
      Separator string to append to each token.  If a 2-element tuple, then
      indicates the separator between tokens and the separator after the last
      token.  Trailing whitespace is removed from each line before appending
      the suffix, but not from between tokens on the same line.
    :param prefix:
      String to prepend at the beginning of each line.  If a 2-element tuple,
      then indicates the prefix for the first line and prefix for subsequent
      lines.
    :param suffix:
      String to append to the end of each line.  If a 2-element tuple, then
      indicates the suffix for all lines except the last, and the suffix for
      the last line.
    :return:
      Filled string.
    """
    N = max_line_length
    assert len(tokens) > 0
    if isinstance(prefix, tuple):
        first_prefix, cont_prefix = prefix
    else:
        first_prefix = cont_prefix = prefix
    if isinstance(suffix, tuple):
        nonterm_suffix, term_suffix = suffix
    else:
        nonterm_suffix = term_suffix = suffix
    if isinstance(sep, tuple):
        nonterm_sep, term_sep = sep
    else:
        nonterm_sep = term_sep = sep
    lines = [first_prefix + tokens[0]]
    for token, is_last in zip(tokens[1:], [False]*(len(tokens)-2) + [True]):
        suffix = term_suffix if is_last else nonterm_suffix
        sep = (term_sep if is_last else nonterm_sep).rstrip()
        # Does the next token fit?
        if len(lines[-1] + nonterm_sep + token + sep + suffix) <= N:
            # Yes; add it.
            lines[-1] += nonterm_sep + token
        else:
            # No; break into new line.
            lines[-1] += nonterm_sep.rstrip() + nonterm_suffix + newline
            lines.append(cont_prefix + token)
    lines[-1] += term_sep.rstrip() + term_suffix + newline
    return ''.join(lines)


def pyfill(prefix, tokens, params=FormatParams()):
    """
    Fill a Python statement.

      >>> print(pyfill('print ', ["foo.bar", "baz", "quux", "quuuuux"]), end='')
      print foo.bar, baz, quux, quuuuux
      >>> print(pyfill('print ', ["foo.bar", "baz", "quux", "quuuuux"],
      ...        FormatParams(max_line_length=15, hanging_indent='auto')), end='')
      print (foo.bar,
             baz,
             quux,
             quuuuux)
      >>> print(pyfill('print ', ["foo.bar", "baz", "quux", "quuuuux"],
      ...        FormatParams(max_line_length=14, hanging_indent='auto')), end='')
      print (
          foo.bar,
          baz, quux,
          quuuuux)

    :param prefix:
      Prefix for first line.
    :param tokens:
      Sequence of string tokens
    :type params:
      `FormatParams`
    :rtype:
      ``str``
    """
    N = params.max_line_length or params._max_line_lenght_default
    if params.wrap_paren:
        # Check how we will break up the tokens.
        len_full = sum(len(tok) for tok in tokens) + 2 * (len(tokens)-1)
        if len(prefix) + len_full <= N:
            # The entire thing fits on one line; no parens needed.  We check
            # this first because breaking into lines adds paren overhead.
            #
            # Output looks like:
            #   from foo import abc, defgh, ijkl, mnopq, rst
            return prefix + ", ".join(tokens) + "\n"
        if params.hanging_indent == "never":
            hanging_indent = False
        elif params.hanging_indent == "always":
            hanging_indent = True
        elif params.hanging_indent == "auto":
            # Decide automatically whether to do hanging-indent mode.  If any
            # line would exceed the max_line_length, then do hanging indent;
            # else don't.
            #
            # In order to use non-hanging-indent mode, the first line would
            # have an overhead of 2 because of "(" and ",".  We check the
            # longest token since even if the first token fits, we still want
            # to avoid later tokens running over N.
            maxtoklen = max(len(token) for token in tokens)
            hanging_indent = (len(prefix) + maxtoklen + 2 > N)
        else:
            raise ValueError("bad params.hanging_indent=%r"
                             % (params.hanging_indent,))
        if hanging_indent:
            # Hanging indent mode.  We need a single opening paren and
            # continue all imports on separate lines.
            #
            # Output looks like:
            #   from foo import (
            #       abc, defgh, ijkl,
            #       mnopq, rst)
            return (prefix + "(\n"
                    + fill(tokens, max_line_length=N,
                           prefix=(" " * params.indent), suffix=("", ")")))
        else:
            # Non-hanging-indent mode.
            #
            # Output looks like:
            #   from foo import (abc, defgh,
            #                    ijkl, mnopq,
            #                    rst)
            pprefix = prefix + "("
            return fill(tokens, max_line_length=N,
                        prefix=(pprefix, " " * len(pprefix)), suffix=("", ")"))
    else:
        raise NotImplementedError
